@@ -22,15 +22,19 @@
  * @short_description: Base for things in a #KgxPages
  */
 
+#include "kgx-config.h"
+
 #include <glib/gi18n.h>
 #define PCRE2_CODE_UNIT_WIDTH 0
 #include <pcre2.h>
 
-#include "kgx-config.h"
 #include "kgx-tab.h"
 #include "kgx-pages.h"
 #include "kgx-terminal.h"
+#include "kgx-settings.h"
+#include "kgx-util.h"
 #include "kgx-application.h"
+#include "kgx-marshals.h"
 
 
 typedef struct _KgxTabPrivate KgxTabPrivate;
@@ -38,16 +42,12 @@ struct _KgxTabPrivate {
   guint                 id;
 
   KgxApplication       *application;
+  KgxSettings          *settings;
 
   char                 *title;
   char                 *tooltip;
   GFile                *path;
-  PangoFontDescription *font;
-  double                zoom;
   KgxStatus             status;
-  KgxTheme              theme;
-  gboolean              opaque;
-  gint64                scrollback_lines;
 
   gboolean              is_active;
   gboolean              close_on_quit;
@@ -55,19 +55,8 @@ struct _KgxTabPrivate {
   gboolean              search_mode_enabled;
 
   KgxTerminal          *terminal;
-  GBinding             *term_title_bind;
-  GBinding             *term_path_bind;
-  GBinding             *term_font_bind;
-  GBinding             *term_zoom_bind;
-  GBinding             *term_theme_bind;
-  GBinding             *term_opaque_bind;
-  GBinding             *term_scrollback_bind;
-
-  GBinding             *pages_font_bind;
-  GBinding             *pages_zoom_bind;
-  GBinding             *pages_theme_bind;
-  GBinding             *pages_opaque_bind;
-  GBinding             *pages_scrollback_bind;
+  GSignalGroup         *terminal_signals;
+  GBindingGroup        *terminal_binds;
 
   GtkWidget            *stack;
   GtkWidget            *spinner_revealer;
@@ -100,19 +89,16 @@ G_DEFINE_ABSTRACT_TYPE_WITH_CODE (KgxTab, kgx_tab, GTK_TYPE_BOX,
 enum {
   PROP_0,
   PROP_APPLICATION,
+  PROP_SETTINGS,
+  PROP_TERMINAL,
   PROP_TAB_TITLE,
   PROP_TAB_PATH,
   PROP_TAB_STATUS,
   PROP_TAB_TOOLTIP,
-  PROP_FONT,
-  PROP_ZOOM,
-  PROP_THEME,
   PROP_IS_ACTIVE,
-  PROP_OPAQUE,
   PROP_CLOSE_ON_QUIT,
   PROP_NEEDS_ATTENTION,
   PROP_SEARCH_MODE_ENABLED,
-  PROP_SCROLLBACK_LINES,
   LAST_PROP
 };
 static GParamSpec *pspecs[LAST_PROP] = { NULL, };
@@ -127,32 +113,6 @@ enum {
 static guint signals[N_SIGNALS];
 
 
-static void
-size_changed (KgxTerminal *term,
-              guint        rows,
-              guint        cols,
-              KgxTab      *self)
-{
-  g_signal_emit (self, signals[SIZE_CHANGED], 0, rows, cols);
-}
-
-
-static void
-font_increase (KgxTerminal *term,
-               KgxTab      *self)
-{
-  g_signal_emit (self, signals[ZOOM], 0, KGX_ZOOM_IN);
-}
-
-
-static void
-font_decrease (KgxTerminal *term,
-               KgxTab      *self)
-{
-  g_signal_emit (self, signals[ZOOM], 0, KGX_ZOOM_OUT);
-}
-
-
 static void
 kgx_tab_dispose (GObject *object)
 {
@@ -168,20 +128,12 @@ kgx_tab_dispose (GObject *object)
   }
 
   g_clear_object (&priv->application);
+  g_clear_object (&priv->settings);
+  g_clear_object (&priv->terminal);
 
   g_clear_pointer (&priv->title, g_free);
   g_clear_pointer (&priv->tooltip, g_free);
   g_clear_object (&priv->path);
-  g_clear_pointer (&priv->font, pango_font_description_free);
-
-  if (priv->terminal) {
-    g_object_disconnect (priv->terminal,
-                        "any-signal::size-changed", G_CALLBACK (size_changed), self,
-                        "any-signal::increase-font-size", G_CALLBACK (font_increase), self,
-                        "any-signal::decrease-font-size", G_CALLBACK (font_decrease), self,
-                        NULL);
-  }
-  g_clear_object (&priv->terminal);
 
   g_clear_pointer (&priv->root, g_hash_table_unref);
   g_clear_pointer (&priv->remote, g_hash_table_unref);
@@ -200,14 +152,14 @@ search_enabled (GObject    *object,
 {
   KgxTabPrivate *priv = kgx_tab_get_instance_private (self);
 
-  if (!hdy_search_bar_get_search_mode (HDY_SEARCH_BAR (priv->search_bar))) {
+  if (!gtk_search_bar_get_search_mode (GTK_SEARCH_BAR (priv->search_bar))) {
     gtk_widget_grab_focus (GTK_WIDGET (self));
   }
 }
 
 
 static void
-search_changed (HdySearchBar *bar,
+search_changed (GtkSearchBar *bar,
                 KgxTab       *self)
 {
   KgxTabPrivate *priv = kgx_tab_get_instance_private (self);
@@ -217,7 +169,7 @@ search_changed (HdySearchBar *bar,
   gboolean narrowing_down;
   guint32 flags = PCRE2_MULTILINE;
 
-  search = gtk_entry_get_text (GTK_ENTRY (priv->search_entry));
+  search = gtk_editable_get_text (GTK_EDITABLE (priv->search_entry));
 
   if (search) {
     g_autofree char *lowercase = g_utf8_strdown (search, -1);
@@ -270,7 +222,7 @@ search_changed (HdySearchBar *bar,
 
 
 static void
-search_next (HdySearchBar *bar,
+search_next (GtkSearchBar *bar,
              KgxTab       *self)
 {
   KgxTabPrivate *priv = kgx_tab_get_instance_private (self);
@@ -280,7 +232,7 @@ search_next (HdySearchBar *bar,
 
 
 static void
-search_prev (HdySearchBar *bar,
+search_prev (GtkSearchBar *bar,
              KgxTab       *self)
 {
   KgxTabPrivate *priv = kgx_tab_get_instance_private (self);
@@ -317,6 +269,34 @@ start_spinner_timeout_cb (KgxTab *self)
 }
 
 
+static void
+set_status (KgxTab    *self,
+            KgxStatus  status)
+{
+  KgxTabPrivate *priv = kgx_tab_get_instance_private (self);
+
+  if (priv->status == status) {
+    return;
+  }
+
+  priv->status = status;
+
+  if (status & KGX_REMOTE) {
+    gtk_widget_add_css_class (GTK_WIDGET (self), KGX_WINDOW_STYLE_REMOTE);
+  } else {
+    gtk_widget_remove_css_class (GTK_WIDGET (self), KGX_WINDOW_STYLE_REMOTE);
+  }
+
+  if (status & KGX_PRIVILEGED) {
+    gtk_widget_add_css_class (GTK_WIDGET (self), KGX_WINDOW_STYLE_ROOT);
+  } else {
+    gtk_widget_remove_css_class (GTK_WIDGET (self), KGX_WINDOW_STYLE_ROOT);
+  }
+
+  g_object_notify_by_pspec (G_OBJECT (self), pspecs[PROP_TAB_STATUS]);
+}
+
+
 static void
 kgx_tab_get_property (GObject    *object,
                       guint       property_id,
@@ -330,6 +310,12 @@ kgx_tab_get_property (GObject    *object,
     case PROP_APPLICATION:
       g_value_set_object (value, priv->application);
       break;
+    case PROP_SETTINGS:
+      g_value_set_object (value, priv->settings);
+      break;
+    case PROP_TERMINAL:
+      g_value_set_object (value, priv->terminal);
+      break;
     case PROP_TAB_TITLE:
       g_value_set_string (value, priv->title);
       break;
@@ -342,21 +328,9 @@ kgx_tab_get_property (GObject    *object,
     case PROP_TAB_TOOLTIP:
       g_value_set_string (value, priv->tooltip);
       break;
-    case PROP_FONT:
-      g_value_set_boxed (value, priv->font);
-      break;
-    case PROP_ZOOM:
-      g_value_set_double (value, priv->zoom);
-      break;
     case PROP_IS_ACTIVE:
       g_value_set_boolean (value, priv->is_active);
       break;
-    case PROP_THEME:
-      g_value_set_enum (value, priv->theme);
-      break;
-    case PROP_OPAQUE:
-      g_value_set_boolean (value, priv->opaque);
-      break;
     case PROP_CLOSE_ON_QUIT:
       g_value_set_boolean (value, priv->close_on_quit);
       break;
@@ -366,9 +340,6 @@ kgx_tab_get_property (GObject    *object,
     case PROP_SEARCH_MODE_ENABLED:
       g_value_set_boolean (value, priv->search_mode_enabled);
       break;
-    case PROP_SCROLLBACK_LINES:
-      g_value_set_int64 (value, priv->scrollback_lines);
-      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
       break;
@@ -420,39 +391,29 @@ kgx_tab_set_property (GObject      *object,
       priv->application = g_value_dup_object (value);
       kgx_application_add_page (priv->application, self);
       break;
+    case PROP_SETTINGS:
+      g_set_object (&priv->settings, g_value_get_object (value));
+      break;
+    case PROP_TERMINAL:
+      g_set_object (&priv->terminal, g_value_get_object (value));
+      break;
     case PROP_TAB_TITLE:
       g_clear_pointer (&priv->title, g_free);
       priv->title = g_value_dup_string (value);
       break;
     case PROP_TAB_PATH:
-      g_clear_object (&priv->path);
-      priv->path = g_value_dup_object (value);
+      g_set_object (&priv->path, g_value_get_object (value));
       break;
     case PROP_TAB_STATUS:
-      priv->status = g_value_get_flags (value);
+      set_status (self, g_value_get_flags (value));
       break;
     case PROP_TAB_TOOLTIP:
       g_clear_pointer (&priv->tooltip, g_free);
       priv->tooltip = g_value_dup_string (value);
       break;
-    case PROP_FONT:
-      if (priv->font) {
-        g_boxed_free (PANGO_TYPE_FONT_DESCRIPTION, priv->font);
-      }
-      priv->font = g_value_dup_boxed (value);
-      break;
-    case PROP_ZOOM:
-      priv->zoom = g_value_get_double (value);
-      break;
     case PROP_IS_ACTIVE:
       kgx_tab_set_is_active (self, g_value_get_boolean (value));
       break;
-    case PROP_THEME:
-      priv->theme = g_value_get_enum (value);
-      break;
-    case PROP_OPAQUE:
-      priv->opaque = g_value_get_boolean (value);
-      break;
     case PROP_CLOSE_ON_QUIT:
       priv->close_on_quit = g_value_get_boolean (value);
       break;
@@ -462,9 +423,6 @@ kgx_tab_set_property (GObject      *object,
     case PROP_SEARCH_MODE_ENABLED:
       priv->search_mode_enabled = g_value_get_boolean (value);
       break;
-    case PROP_SCROLLBACK_LINES:
-      priv->scrollback_lines = g_value_get_int64 (value);
-      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
       break;
@@ -473,48 +431,17 @@ kgx_tab_set_property (GObject      *object,
 
 
 static gboolean
-kgx_tab_draw (GtkWidget *widget,
-              cairo_t   *cr)
+drop (GtkDropTarget *target,
+      const GValue  *value,
+      KgxTab        *self)
 {
-  GtkStateFlags flags;
+  kgx_tab_accept_drop (self, value);
 
-  GTK_WIDGET_CLASS (kgx_tab_parent_class)->draw (widget, cr);
-
-  flags = gtk_widget_get_state_flags (widget);
-
-  /* FIXME this is a workaround for the fact GTK3 css outline is used for focus
-   * only by default. This can be removed in GTK4 as it just works there. */
-  if (flags & GTK_STATE_FLAG_DROP_ACTIVE) {
-    GtkStyleContext *context = gtk_widget_get_style_context (widget);
-    int width = gtk_widget_get_allocated_width (widget);
-    int height = gtk_widget_get_allocated_height (widget);
-
-    gtk_render_focus (context, cr, 0, 0, width, height);
-  }
-
-  return GDK_EVENT_PROPAGATE;
+  return TRUE;
 }
 
 
-static void
-kgx_tab_drag_data_received (GtkWidget        *widget,
-                            GdkDragContext   *context,
-                            gint              x,
-                            gint              y,
-                            GtkSelectionData *selection_data,
-                            guint             info,
-                            guint             time)
-{
-  KgxTab *self = KGX_TAB (widget);
-  GdkDragAction action = gdk_drag_context_get_selected_action (context);
-
-  kgx_tab_accept_drop (self, selection_data);
-
-  gtk_drag_finish (context, TRUE, action == GDK_ACTION_COPY, time);
-}
-
-
-static void
+static gboolean
 kgx_tab_grab_focus (GtkWidget *widget)
 {
   KgxTab *self = KGX_TAB (widget);
@@ -522,10 +449,10 @@ kgx_tab_grab_focus (GtkWidget *widget)
 
   if (priv->terminal) {
     gtk_widget_grab_focus (GTK_WIDGET (priv->terminal));
-    return;
+    return GDK_EVENT_STOP;
   }
 
-  GTK_WIDGET_CLASS (kgx_tab_parent_class)->grab_focus (widget);
+  return GTK_WIDGET_CLASS (kgx_tab_parent_class)->grab_focus (widget);
 }
 
 
@@ -549,6 +476,30 @@ kgx_tab_real_start_finish (KgxTab        *tab,
 }
 
 
+static void
+kgx_tab_real_died (KgxTab         *self,
+                   GtkMessageType  type,
+                   const char     *message,
+                   gboolean        success)
+{
+  KgxTabPrivate *priv;
+
+  g_return_if_fail (KGX_IS_TAB (self));
+
+  priv = kgx_tab_get_instance_private (self);
+
+  gtk_label_set_markup (GTK_LABEL (priv->label), message);
+
+  if (type == GTK_MESSAGE_ERROR) {
+    gtk_widget_add_css_class (priv->revealer, "error");
+  } else {
+    gtk_widget_remove_css_class (priv->revealer, "error");
+  }
+
+  gtk_revealer_set_reveal_child (GTK_REVEALER (priv->revealer), TRUE);
+}
+
+
 static void
 kgx_tab_class_init (KgxTabClass *klass)
 {
@@ -560,12 +511,11 @@ kgx_tab_class_init (KgxTabClass *klass)
   object_class->get_property = kgx_tab_get_property;
   object_class->set_property = kgx_tab_set_property;
 
-  widget_class->draw = kgx_tab_draw;
-  widget_class->drag_data_received = kgx_tab_drag_data_received;
   widget_class->grab_focus = kgx_tab_grab_focus;
 
   tab_class->start = kgx_tab_real_start;
   tab_class->start_finish = kgx_tab_real_start_finish;
+  tab_class->died = kgx_tab_real_died;
 
   /**
    * KgxTab:application:
@@ -579,6 +529,16 @@ kgx_tab_class_init (KgxTabClass *klass)
                          KGX_TYPE_APPLICATION,
                          G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
 
+  pspecs[PROP_SETTINGS] =
+    g_param_spec_object ("settings", NULL, NULL,
+                         KGX_TYPE_SETTINGS,
+                         G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+
+  pspecs[PROP_TERMINAL] =
+    g_param_spec_object ("terminal", NULL, NULL,
+                         KGX_TYPE_TERMINAL,
+                         G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+
   /**
    * KgxTab:tab-title:
    *
@@ -620,18 +580,6 @@ kgx_tab_class_init (KgxTabClass *klass)
                          NULL,
                          G_PARAM_READWRITE);
 
-  pspecs[PROP_FONT] =
-    g_param_spec_boxed ("font", "Font", "Monospace font",
-                         PANGO_TYPE_FONT_DESCRIPTION,
-                         G_PARAM_READWRITE);
-
-  pspecs[PROP_ZOOM] =
-    g_param_spec_double ("zoom", "Zoom", "Font scaling",
-                         KGX_FONT_SCALE_MIN,
-                         KGX_FONT_SCALE_MAX,
-                         KGX_FONT_SCALE_DEFAULT,
-                         G_PARAM_READWRITE);
-
   /**
    * KgxTab:is-active:
    *
@@ -644,33 +592,6 @@ kgx_tab_class_init (KgxTabClass *klass)
                           FALSE,
                           G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY);
 
-  /**
-   * KgxTab:theme:
-   *
-   * The #KgxTheme to apply to the #KgxTerminal s in the #KgxTab
-   *
-   * Stability: Private
-   */
-  pspecs[PROP_THEME] =
-    g_param_spec_enum ("theme", "Theme", "The path of the active tab",
-                       KGX_TYPE_THEME,
-                       KGX_THEME_NIGHT,
-                       G_PARAM_READWRITE);
-
-  /**
-   * KgxTab:opaque:
-   *
-   * Whether to disable transparency
-   *
-   * Bound to #GtkWindow:is-maximized on the #KgxWindow
-   *
-   * Stability: Private
-   */
-  pspecs[PROP_OPAQUE] =
-    g_param_spec_boolean ("opaque", "Opaque", "Terminal opaqueness",
-                          FALSE,
-                          G_PARAM_READWRITE);
-
   pspecs[PROP_CLOSE_ON_QUIT] =
     g_param_spec_boolean ("close-on-quit", "Close on quit",
                           "Should the tab close when dead",
@@ -689,26 +610,13 @@ kgx_tab_class_init (KgxTabClass *klass)
                           FALSE,
                           G_PARAM_READWRITE);
 
-  /**
-   * KgxTab:scrollback-lines:
-   *
-   * How many lines of scrollback #KgxTerminal should keep
-   *
-   * Bound to ‘scrollback-lines’ GSetting so changes persist
-   *
-   * Stability: Private
-   */
-  pspecs[PROP_SCROLLBACK_LINES] =
-    g_param_spec_int64 ("scrollback-lines", "Scrollback Lines", "Size of the scrollback",
-                        G_MININT64, G_MAXINT64, 512,
-                        G_PARAM_READWRITE);
-
   g_object_class_install_properties (object_class, LAST_PROP, pspecs);
 
   signals[SIZE_CHANGED] = g_signal_new ("size-changed",
                                         G_TYPE_FROM_CLASS (klass),
                                         G_SIGNAL_RUN_LAST,
-                                        0, NULL, NULL, NULL,
+                                        0, NULL, NULL,
+                                        kgx_marshals_VOID__UINT_UINT,
                                         G_TYPE_NONE,
                                         2,
                                         G_TYPE_UINT,
@@ -717,7 +625,8 @@ kgx_tab_class_init (KgxTabClass *klass)
   signals[ZOOM] = g_signal_new ("zoom",
                                 G_TYPE_FROM_CLASS (klass),
                                 G_SIGNAL_RUN_LAST,
-                                0, NULL, NULL, NULL,
+                                0, NULL, NULL,
+                                kgx_marshals_VOID__ENUM,
                                 G_TYPE_NONE,
                                 1,
                                 KGX_TYPE_ZOOM);
@@ -725,7 +634,9 @@ kgx_tab_class_init (KgxTabClass *klass)
   signals[DIED] = g_signal_new ("died",
                                 G_TYPE_FROM_CLASS (klass),
                                 G_SIGNAL_RUN_LAST,
-                                0, NULL, NULL, NULL,
+                                G_STRUCT_OFFSET (KgxTabClass, died),
+                                NULL, NULL,
+                                kgx_marshals_VOID__ENUM_STRING_BOOLEAN,
                                 G_TYPE_NONE,
                                 3,
                                 GTK_TYPE_MESSAGE_TYPE,
@@ -741,6 +652,8 @@ kgx_tab_class_init (KgxTabClass *klass)
   gtk_widget_class_bind_template_child_private (widget_class, KgxTab, label);
   gtk_widget_class_bind_template_child_private (widget_class, KgxTab, search_entry);
   gtk_widget_class_bind_template_child_private (widget_class, KgxTab, search_bar);
+  gtk_widget_class_bind_template_child_private (widget_class, KgxTab, terminal_signals);
+  gtk_widget_class_bind_template_child_private (widget_class, KgxTab, terminal_binds);
 
   gtk_widget_class_bind_template_callback (widget_class, search_enabled);
   gtk_widget_class_bind_template_callback (widget_class, search_changed);
@@ -768,8 +681,8 @@ kgx_tab_add_child (GtkBuildable *buildable,
   if (type && g_str_equal (type, "content")) {
     g_set_weak_pointer (&priv->content, GTK_WIDGET (child));
     gtk_stack_add_named (GTK_STACK (priv->stack), GTK_WIDGET (child), "content");
-  } else {
-    gtk_container_add (GTK_CONTAINER (self), GTK_WIDGET (child));
+  } else if (GTK_IS_WIDGET (child)) {
+    gtk_box_append (GTK_BOX (self), GTK_WIDGET (child));
   }
 }
 
@@ -782,33 +695,28 @@ kgx_tab_buildable_iface_init (GtkBuildableIface *iface)
 
 
 static void
-died (KgxTab         *self,
-      GtkMessageType  type,
-      const char     *message,
-      gboolean        success)
+size_changed (KgxTerminal *term,
+              guint        rows,
+              guint        cols,
+              KgxTab      *self)
 {
-  KgxTabPrivate *priv;
-  GtkStyleContext *context;
+  g_signal_emit (self, signals[SIZE_CHANGED], 0, rows, cols);
+}
 
-  g_return_if_fail (KGX_IS_TAB (self));
 
-  priv = kgx_tab_get_instance_private (self);
+static void
+font_increase (KgxTerminal *term,
+               KgxTab      *self)
+{
+  g_signal_emit (self, signals[ZOOM], 0, KGX_ZOOM_IN);
+}
 
-  gtk_label_set_markup (GTK_LABEL (priv->label), message);
 
-  context = gtk_widget_get_style_context (GTK_WIDGET (priv->revealer));
-
-  if (type == GTK_MESSAGE_ERROR) {
-    gtk_style_context_add_class (context, "error");
-  } else {
-    gtk_style_context_remove_class (context, "error");
-  }
-
-  gtk_revealer_set_reveal_child (GTK_REVEALER (priv->revealer), TRUE);
-
-  if (priv->close_on_quit && success) {
-    kgx_pages_remove_page (kgx_tab_get_pages (self), self);
-  }
+static void
+font_decrease (KgxTerminal *term,
+               KgxTab      *self)
+{
+  g_signal_emit (self, signals[ZOOM], 0, KGX_ZOOM_OUT);
 }
 
 
@@ -817,6 +725,7 @@ kgx_tab_init (KgxTab *self)
 {
   static guint last_id = 0;
   KgxTabPrivate *priv = kgx_tab_get_instance_private (self);
+  GtkDropTarget *target;
 
   last_id++;
 
@@ -831,116 +740,29 @@ kgx_tab_init (KgxTab *self)
 
   gtk_widget_init_template (GTK_WIDGET (self));
 
-  g_signal_connect (self, "died", G_CALLBACK (died), NULL);
+  g_signal_group_connect (priv->terminal_signals,
+                          "size-changed", G_CALLBACK (size_changed),
+                          self),
+  g_signal_group_connect (priv->terminal_signals,
+                          "increase-font-size", G_CALLBACK (font_increase),
+                          self),
+  g_signal_group_connect (priv->terminal_signals,
+                          "decrease-font-size", G_CALLBACK (font_decrease),
+                          self),
 
-  hdy_search_bar_connect_entry (HDY_SEARCH_BAR (priv->search_bar),
-                                GTK_ENTRY (priv->search_entry));
+  g_binding_group_bind (priv->terminal_binds, "window-title",
+                        self, "tab-title",
+                        G_BINDING_SYNC_CREATE);
+  g_binding_group_bind (priv->terminal_binds, "path",
+                        self, "tab-path",
+                        G_BINDING_SYNC_CREATE);
 
-  gtk_drag_dest_set (GTK_WIDGET (self), GTK_DEST_DEFAULT_ALL, NULL, 0,
-                     GDK_ACTION_COPY);
-  gtk_drag_dest_add_text_targets (GTK_WIDGET (self));
-}
+  gtk_search_bar_connect_entry (GTK_SEARCH_BAR (priv->search_bar),
+                                GTK_EDITABLE (priv->search_entry));
 
-
-void
-kgx_tab_connect_terminal (KgxTab      *self,
-                          KgxTerminal *term)
-{
-  KgxTabPrivate *priv;
-
-  g_return_if_fail (KGX_IS_TAB (self));
-  g_return_if_fail (KGX_IS_TERMINAL (term));
-
-  priv = kgx_tab_get_instance_private (self);
-
-  if (priv->terminal == term) {
-    return;
-  }
-
-  if (priv->terminal) {
-    g_object_disconnect (priv->terminal,
-                         "any-signal::size-changed", G_CALLBACK (size_changed), self,
-                         "any-signal::increase-font-size", G_CALLBACK (font_increase), self,
-                         "any-signal::decrease-font-size", G_CALLBACK (font_decrease), self,
-                         NULL);
-  }
-
-  g_clear_object (&priv->term_title_bind);
-  g_clear_object (&priv->term_path_bind);
-  g_clear_object (&priv->term_font_bind);
-  g_clear_object (&priv->term_zoom_bind);
-  g_clear_object (&priv->term_theme_bind);
-  g_clear_object (&priv->term_opaque_bind);
-  g_clear_object (&priv->term_scrollback_bind);
-
-  g_set_object (&priv->terminal, term);
-
-  g_object_connect (term,
-                    "signal::size-changed", G_CALLBACK (size_changed), self,
-                    "signal::increase-font-size", G_CALLBACK (font_increase), self,
-                    "signal::decrease-font-size", G_CALLBACK (font_decrease), self,
-                    NULL);
-
-  priv->term_title_bind = g_object_bind_property (term, "window-title",
-                                                  self, "tab-title",
-                                                  G_BINDING_SYNC_CREATE);
-  priv->term_path_bind = g_object_bind_property (term, "path",
-                                                 self, "tab-path",
-                                                 G_BINDING_SYNC_CREATE);
-  priv->term_font_bind = g_object_bind_property (self, "font",
-                                                 term, "font-desc",
-                                                 G_BINDING_SYNC_CREATE);
-  priv->term_zoom_bind = g_object_bind_property (self, "zoom",
-                                                 term, "font-scale",
-                                                 G_BINDING_SYNC_CREATE);
-  priv->term_theme_bind = g_object_bind_property (self, "theme",
-                                                  term, "theme",
-                                                  G_BINDING_SYNC_CREATE);
-  priv->term_opaque_bind = g_object_bind_property (self, "opaque",
-                                                   term, "opaque",
-                                                   G_BINDING_SYNC_CREATE);
-  priv->term_scrollback_bind = g_object_bind_property (self, "scrollback-lines",
-                                                       term, "scrollback-lines",
-                                                       G_BINDING_SYNC_CREATE);
-}
-
-
-void
-kgx_tab_set_pages (KgxTab   *self,
-                   KgxPages *pages)
-{
-  KgxTabPrivate *priv;
-
-  g_return_if_fail (KGX_IS_TAB (self));
-  g_return_if_fail (KGX_IS_PAGES (pages) || !pages);
-
-  priv = kgx_tab_get_instance_private (self);
-
-  g_clear_object (&priv->pages_font_bind);
-  g_clear_object (&priv->pages_zoom_bind);
-  g_clear_object (&priv->pages_theme_bind);
-  g_clear_object (&priv->pages_opaque_bind);
-  g_clear_object (&priv->pages_scrollback_bind);
-
-  if (pages == NULL) {
-    return;
-  }
-
-  priv->pages_font_bind = g_object_bind_property (pages, "font",
-                                                  self, "font",
-                                                  G_BINDING_SYNC_CREATE);
-  priv->pages_zoom_bind = g_object_bind_property (pages, "zoom",
-                                                  self, "zoom",
-                                                  G_BINDING_SYNC_CREATE);
-  priv->pages_theme_bind = g_object_bind_property (pages, "theme",
-                                                   self, "theme",
-                                                   G_BINDING_SYNC_CREATE);
-  priv->pages_opaque_bind = g_object_bind_property (pages, "opaque",
-                                                    self, "opaque",
-                                                    G_BINDING_SYNC_CREATE);
-  priv->pages_scrollback_bind = g_object_bind_property (pages, "scrollback-lines",
-                                                        self, "scrollback-lines",
-                                                        G_BINDING_SYNC_CREATE);
+  target = gtk_drop_target_new (G_TYPE_STRING, GDK_ACTION_COPY);
+  g_signal_connect (target, "drop", G_CALLBACK (drop), self);
+  gtk_widget_add_controller (GTK_WIDGET (self), GTK_EVENT_CONTROLLER (target));
 }
 
 
@@ -1069,7 +891,8 @@ kgx_tab_push_child (KgxTab     *self,
 {
   GtkStyleContext *context;
   GPid pid = 0;
-  const char *exec = NULL;
+  GStrv argv;
+  g_autofree char *program = NULL;
   KgxStatus new_status = KGX_NONE;
   KgxTabPrivate *priv;
 
@@ -1079,22 +902,35 @@ kgx_tab_push_child (KgxTab     *self,
 
   context = gtk_widget_get_style_context (GTK_WIDGET (self));
   pid = kgx_process_get_pid (process);
-  exec = kgx_process_get_exec (process);
+  argv = kgx_process_get_argv (process);
 
-  if (G_UNLIKELY (g_str_has_prefix (exec, "ssh "))) {
+  if (G_LIKELY (argv[0] != NULL)) {
+    program = g_path_get_basename (argv[0]);
+  }
+
+  if (G_UNLIKELY (g_strcmp0 (program, "ssh") == 0 ||
+                  g_strcmp0 (program, "mosh-client") == 0 ||
+                  g_strcmp0 (program, "mosh") == 0 ||
+                  g_strcmp0 (program, "et") == 0)) {
     new_status |= push_type (priv->remote, pid, NULL, context, KGX_REMOTE);
   }
 
+  if (G_UNLIKELY (g_strcmp0 (program, "waypipe") == 0)) {
+    for (int i = 1; argv[i]; i++) {
+      if (G_UNLIKELY (g_strcmp0 (argv[i], "ssh") == 0)) {
+        new_status |= push_type (priv->remote, pid, NULL, context, KGX_REMOTE);
+        break;
+      }
+    }
+  }
+
   if (G_UNLIKELY (kgx_process_get_is_root (process))) {
     new_status |= push_type (priv->root, pid, NULL, context, KGX_PRIVILEGED);
   }
 
   push_type (priv->children, pid, process, context, KGX_NONE);
 
-  if (priv->status != new_status) {
-    priv->status = new_status;
-    g_object_notify_by_pspec (G_OBJECT (self), pspecs[PROP_TAB_STATUS]);
-  }
+  set_status (self, new_status);
 }
 
 
@@ -1149,10 +985,7 @@ kgx_tab_pop_child (KgxTab     *self,
   new_status |= pop_type (priv->root, pid, context, KGX_PRIVILEGED);
   pop_type (priv->children, pid, context, KGX_NONE);
 
-  if (priv->status != new_status) {
-    priv->status = new_status;
-    g_object_notify_by_pspec (G_OBJECT (self), pspecs[PROP_TAB_STATUS]);
-  }
+  set_status (self, new_status);
 
   if (!kgx_tab_is_active (self)) {
     g_autoptr (GNotification) noti = NULL;
@@ -1224,23 +1057,26 @@ kgx_tab_get_children (KgxTab *self)
 
 
 void
-kgx_tab_accept_drop (KgxTab           *self,
-                     GtkSelectionData *selection_data)
+kgx_tab_accept_drop (KgxTab       *self,
+                     const GValue *value)
 {
   KgxTabPrivate *priv;
   g_autofree char *text = NULL;
+  g_auto (GStrv) uris = NULL;
 
   g_return_if_fail (KGX_IS_TAB (self));
 
   priv = kgx_tab_get_instance_private (self);
 
-  if (gtk_selection_data_get_length (selection_data) < 0)
-    return;
+  uris = g_strsplit (g_value_get_string (value), "\n", 0);
 
-  text = (char *) gtk_selection_data_get_text (selection_data);
+  kgx_util_transform_uris_to_quoted_fuse_paths (uris);
 
-  if (priv->terminal)
+  text = kgx_util_concat_uris (uris, NULL);
+
+  if (priv->terminal) {
     kgx_terminal_accept_paste (KGX_TERMINAL (priv->terminal), text);
+  }
 }
 
 
@@ -1263,28 +1099,3 @@ kgx_tab_set_initial_title (KgxTab     *self,
                 "tab-path", path,
                 NULL);
 }
-
-
-gboolean
-kgx_tab_key_press_event (KgxTab   *self,
-                         GdkEvent *event)
-{
-  KgxTabPrivate *priv;
-  GtkWidget *toplevel, *focus;
-
-  g_return_val_if_fail (KGX_IS_TAB (self), GDK_EVENT_PROPAGATE);
-  g_return_val_if_fail (event != NULL, GDK_EVENT_PROPAGATE);
-
-  priv = kgx_tab_get_instance_private (self);
-  toplevel = gtk_widget_get_toplevel (GTK_WIDGET (self));
-
-  if (!GTK_IS_WINDOW (toplevel))
-    return GDK_EVENT_PROPAGATE;
-
-  focus = gtk_window_get_focus (GTK_WINDOW (toplevel));
-
-  if (focus == GTK_WIDGET (priv->terminal))
-    return gtk_widget_event (GTK_WIDGET (priv->terminal), event);
-
-  return GDK_EVENT_PROPAGATE;
-}
